import sys
sys.path.insert(1, '..')

import unittest
import numpy as np
from numpy.testing import assert_array_almost_equal, assert_array_equal
from findiff.diff import *
#import matplotlib.pyplot as plt
import findiff
from numpy import cos, sin
from findiff.grids import UniformGrid


class DiffTest(unittest.TestCase):

    def test_partial_d_dx(self):
        shape = 101,
        x, dx = make_grid(shape, (0, 1))

        u = x**2
        expected = 2 * x

        fd = Diff(0, 1)
        actual = fd(u, dx)

        assert_array_almost_equal(expected, actual)

    def test_partial_d2_dx2(self):
        shape = 101,
        x, dx = make_grid(shape, (0, 1))

        u = x ** 2
        expected = 2

        fd = Diff(0, 2)
        actual = fd(u, dx)

        assert_array_almost_equal(expected, actual)

    def test_partial_d_dx_acc(self):
        shape = 11,
        x, dx = make_grid(shape, (0, 1))

        u = x ** 3
        expected = 3*x**2

        fd = Diff(0, 1)
        actual = fd(u, dx)
        np.testing.assert_raises(AssertionError, assert_array_almost_equal, expected, actual)

        actual = fd(u, dx, acc=4)
        assert_array_almost_equal(expected, actual)

    def test_partial_d2_dxdy(self):

        shape = 50, 50
        (x, y), (dx, dy), (X, Y) = make_grid(shape, ((0, 1), (0, 1)))
        u = np.sin(X) * np.sin(Y)

        fd = Diff(0, 1) * Diff(1, 1)
        actual = fd(u, h={0: dx, 1: dy})

        expected = np.cos(X) * np.cos(Y)
        assert_array_almost_equal(expected, actual, decimal=3)

        fd = Diff(1, 1) * Diff(0, 1)
        actual = fd(u, h={0: dx, 1: dy})
        assert_array_almost_equal(expected, actual, decimal=3)

    def test_partial_d_dx_on_2d_array(self):

        shape = 50, 50
        (x, y), (dx, dy), (X, Y) = make_grid(shape, ((0, 1), (0, 1)))
        u = np.sin(X) * np.sin(Y)

        fd = Diff(1, 1)
        actual = fd(u, dy)

        expected = np.sin(X) * np.cos(Y)
        assert_array_almost_equal(expected, actual, decimal=3)

    def test_add_two_diffs(self):

        shape = 50, 50
        (x, y), (dx, dy), (X, Y) = make_grid(shape, ((0, 1), (0, 1)))
        u = np.sin(X) * np.sin(Y)

        fd = Diff(0, 2) + Diff(1, 2)
        actual = fd(u, dy)

        expected = -2 * np.sin(X) * np.sin(Y)
        assert_array_almost_equal(expected, actual, decimal=3)

    def test_sub_two_diffs(self):

        shape = 50, 50
        (x, y), (dx, dy), (X, Y) = make_grid(shape, ((0, 1), (0, 1)))
        u = np.sin(X) * np.sin(Y)

        fd = Diff(0, 2) - Diff(1, 2)
        actual = fd(u, dy)

        expected = np.zeros(shape)
        assert_array_almost_equal(expected, actual, decimal=3)

    def test_multiply_with_scalar(self):

        shape = 50, 50
        (x, y), (dx, dy), (X, Y) = make_grid(shape, ((0, 1), (0, 1)))
        u = np.sin(X) * np.sin(Y)

        fd = 3 * Diff(0, 2)
        actual = fd(u, dy, acc=4)

        expected = -3 * np.sin(X) * np.sin(Y)
        assert_array_almost_equal(expected, actual)

    def test_multiply_with_variable_from_left(self):

        shape = 50, 50
        (x, y), (dx, dy), (X, Y) = make_grid(shape, ((0, 1), (0, 1)))
        u = np.sin(X) * np.sin(Y)

        fd = Coef(3 * X * Y) * Diff(0, 2)
        actual = fd(u, dy, acc=4)

        expected = - 3*X *Y* np.sin(X) * np.sin(Y)
        assert_array_almost_equal(expected, actual)

    def test_identity(self):

        shape = 50, 50
        (x, y), (dx, dy), (X, Y) = make_grid(shape, ((0, 1), (0, 1)))
        u = np.sin(X) * np.sin(Y)

        fd = Diff(0, 2) + Id()
        actual = fd(u, dy, acc=4)

        expected = np.zeros_like(u)
        assert_array_almost_equal(expected, actual)

    def test_linear_comb(self):
        shape = 50, 50
        (x, y), (dx, dy), (X, Y) = make_grid(shape, ((0, 1), (0, 1)))
        u = np.sin(X) * np.sin(Y)

        fd = Coef(3*X)*Diff(0, 2) - Coef(Y) * (Diff(1, 2) + Diff(0, 1))
        actual = fd(u, h={0: dx, 1:dy}, acc=4)

        expected = -3*X*u - Y * (-sin(X)*sin(Y) + cos(X)*sin(Y))
        assert_array_almost_equal(expected, actual)


def make_grid(shape, edges):

    if len(shape) == 1:
        x = np.linspace(*edges, shape[0])
        dx = x[1]-x[0]
        return x, dx

    axes = tuple([np.linspace(edges[k][0], edges[k][1], shape[k]) for k in range(len(shape))])
    coords = np.meshgrid(*axes, indexing='ij')
    spacings = [axes[k][1]-axes[k][0] for k in range(len(shape))]
    return axes, spacings, coords



import sys
sys.path.insert(1, '..')

import unittest
import numpy as np
from numpy.testing import assert_array_almost_equal, assert_array_equal
from findiff.utils import *

class TestUtils(unittest.TestCase):

    def test_all_index_tuples_for_shape(self):

        shape = 2, 3

        expected = []
        for ix in range(shape[0]):
            for iy in range(shape[1]):
                expected.append((ix, iy))

        actual = all_index_tuples_as_list(shape)

        for a, e in zip(actual, expected):
            self.assertEqual(e, a)

    def test_to_long_index(self):

        shape = 3, 4, 5
        all_tuples = all_index_tuples_as_list(shape)

        expected = list(range(np.prod(shape)))

        for i, idx in enumerate(all_tuples):
            actual = to_long_index(idx, shape)
            self.assertEqual(expected[i], actual)

    def test_to_index_tuple(self):
        shape = 3, 4, 5

        all_tuples = all_index_tuples_as_list(shape)

        for long_idx in range(np.prod(shape)):
            expected = all_tuples[long_idx]
            actual = to_index_tuple(long_idx, shape)
            np.testing.assert_array_equal(expected, actual)




"""A module for the common differential operators of vector calculus"""

import numpy as np
from .operators import FinDiff
from.diff import Diff


class VectorOperator(object):
    """Base class for all vector differential operators.
       Shall not be instantiated directly, but through the child classes.
    """

    def __init__(self, **kwargs):
        """Constructor for the VectorOperator base class.
        
            kwargs:
            -------
            
            h       list with the grid spacings of an N-dimensional uniform grid
            
            coords  list of 1D arrays with the coordinate values along the N axes.
                    This is used for non-uniform grids. 
            
            Either specify "h" or "coords", not both.
        
        """

        if "acc" in kwargs:
            self.acc = kwargs.pop("acc")
        else:
            self.acc = 2

        if "spac" in kwargs or "h" in kwargs: # necessary for backward compatibility 0.5.2 => 0.6
            if "spac" in kwargs:
                kw = "spac"
            else:
                kw = "h"
            self.h = kwargs.pop(kw)
            self.ndims = len(self.h)
            self.components = [FinDiff(k, self.h[k], 1) for k in range(self.ndims)]

        if "coords" in kwargs:
            coords = kwargs.pop("coords")
            self.ndims = self.__get_dimension(coords)
            self.components = [FinDiff((k, coords[k], 1), **kwargs) for k in range(self.ndims)]

    def __get_dimension(self, coords):
        if isinstance(coords, np.ndarray):
            shape = coords.shape
            if len(shape) > 1:
                ndims = shape[0]
            else:
                ndims = 1
        else:
            ndims = len(coords)
        return ndims


class Gradient(VectorOperator):
    r"""
    The N-dimensional gradient.
    
    .. math::
        \nabla = \left(\frac{\partial}{\partial x_0}, \frac{\partial}{\partial x_1}, ... , \frac{\partial}{\partial x_{N-1}}\right)

    :param kwargs:  exactly one of *h* and *coords* must be specified
    
             *h* 
                     list with the grid spacings of an N-dimensional uniform grid     
             *coords*
                     list of 1D arrays with the coordinate values along the N axes.
                     This is used for non-uniform grids.
                     
             *acc*
                     accuracy order, must be positive integer, default is 2
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __call__(self, f):
        """
        Applies the N-dimensional gradient to the array f.
        
        :param f:  ``numpy.ndarray``
        
                Array to apply the gradient to. It represents a scalar function,
                so it must have N axes for the N independent variables.        
           
        :returns: ``numpy.ndarray``
         
                The gradient of f, which has N+1 axes, i.e. it is 
                an array of N arrays of N axes each.
           
        """

        if not isinstance(f, np.ndarray):
            raise TypeError("Function to differentiate must be numpy.ndarray")

        if len(f.shape) != self.ndims:
            raise ValueError("Gradients can only be applied to scalar functions")

        result = []
        for k in range(self.ndims):
            d_dxk = self.components[k]
            result.append(d_dxk(f, acc=self.acc))

        return np.array(result)


class Divergence(VectorOperator):
    r"""
    The N-dimensional divergence.
    
    .. math::
    
       {\rm \bf div} = \nabla \cdot
    
    :param kwargs:  exactly one of *h* and *coords* must be specified

         *h* 
                 list with the grid spacings of an N-dimensional uniform grid     
         *coords*
                 list of 1D arrays with the coordinate values along the N axes.
                 This is used for non-uniform grids.
                 
         *acc*
                 accuracy order, must be positive integer, default is 2
    
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __call__(self, f):
        """
        Applies the divergence to the array f.

        :param f: ``numpy.ndarray``
                
               a vector function of N variables, so its array has N+1 axes.
        
        :returns: ``numpy.ndarray``
            
               the divergence, which is a scalar function of N variables, so it's array dimension has N axes
                
        """
        if not isinstance(f, np.ndarray) and not isinstance(f, list):
            raise TypeError("Function to differentiate must be numpy.ndarray or list of numpy.ndarrays")

        if len(f.shape) != self.ndims + 1 and f.shape[0] != self.ndims:
            raise ValueError("Divergence can only be applied to vector functions of the same dimension")

        result = np.zeros(f.shape[1:])

        for k in range(self.ndims):
            result += self.components[k](f[k], acc=self.acc)

        return result


class Curl(VectorOperator):
    r"""
    The curl operator. 
    
    .. math::
    
        {\rm \bf rot} = \nabla \times
    
    Is only defined for 3D.
    
    :param kwargs:  exactly one of *h* and *coords* must be specified

     *h* 
             list with the grid spacings of a 3-dimensional uniform grid     
     *coords*
             list of 1D arrays with the coordinate values along the 3 axes.
             This is used for non-uniform grids.
             
     *acc*
             accuracy order, must be positive integer, default is 2

    
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if self.ndims != 3:
            raise ValueError("Curl operation is only defined in 3 dimensions. {} were given.".format(self.ndims))

    def __call__(self, f):
        """
        Applies the curl to the array f.

        :param f: ``numpy.ndarray``

               a vector function of N variables, so its array has N+1 axes.

        :returns: ``numpy.ndarray``

               the curl, which is a vector function of N variables, so it's array dimension has N+1 axes

        """

        if not isinstance(f, np.ndarray) and not isinstance(f, list):
            raise TypeError("Function to differentiate must be numpy.ndarray or list of numpy.ndarrays")

        if len(f.shape) != self.ndims + 1 and f.shape[0] != self.ndims:
            raise ValueError("Curl can only be applied to vector functions of the three dimensions")

        result = np.zeros(f.shape)

        result[0] += self.components[1](f[2], acc=self.acc) - self.components[2](f[1], acc=self.acc)
        result[1] += self.components[2](f[0], acc=self.acc) - self.components[0](f[2], acc=self.acc)
        result[2] += self.components[0](f[1], acc=self.acc) - self.components[1](f[0], acc=self.acc)

        return result


class Laplacian(object):
    r"""
        The N-dimensional Laplace operator.

        .. math::

           {\rm \bf \nabla^2} = \sum_{k=0}^{N-1} \frac{\partial^2}{\partial x_k^2}

        :param kwargs:  exactly one of *h* and *coords* must be specified

             *h* 
                     list with the grid spacings of an N-dimensional uniform grid     
             *coords*
                     list of 1D arrays with the coordinate values along the N axes.
                     This is used for non-uniform grids.

             *acc*
                     accuracy order, must be positive integer, default is 2

        """

    """A representation of the Laplace operator in arbitrary dimensions using finite difference schemes"""

    def __init__(self, h=[1.], acc=2):
        h = wrap_in_ndarray(h)

        self._parts = [FinDiff((k, h[k], 2), acc=acc) for k in range(len(h))]

    def __call__(self, f):
        """
        Applies the Laplacian to the array f.

        :param f: ``numpy.ndarray``

               a scalar function of N variables, so its array has N axes.

        :returns: ``numpy.ndarray``

               the Laplacian of f, which is a scalar function of N variables, so it's array dimension has N axes

        """
        laplace_f = np.zeros_like(f)

        for part in self._parts:
            laplace_f += part(f)

        return laplace_f


def wrap_in_ndarray(value):
    """Wraps the argument in a numpy.ndarray.

       If value is a scalar, it is converted in a list first.
       If value is array-like, the shape is conserved.

    """

    if hasattr(value, "__len__"):
        return np.array(value)
    else:
        return np.array([value])



from .coefs import coefficients
from .operators import FinDiff, Coef, Identity, Coefficient
from .vector import Gradient, Divergence, Curl, Laplacian
from .pde import PDE, BoundaryConditions

__version__ = "0.8.6"



from itertools import product
import numpy as np
from .utils import to_long_index, to_index_tuple


class Stencil(object):
    """
    Represent the finite difference stencil for a given differential operator.
    """

    def __init__(self, diff_op, shape, old_stl=None, acc=None):
        """
        Constructor for Stencil objects.

        :param shape: tuple of ints
            Shape of the grid on which the stencil should be applied.

        :param axis: int >= 0
            The coordinate axis along which to take the partial derivative.

        :param order: int > 0
            The order of the derivative.

        :param h: float
            The spacing of the (equidistant) grid

        :param acc: (even) int > 0
            The desired accuracy order of the finite difference scheme.

        """

        self.shape = shape
        self.diff_op = diff_op
        self.char_pts = self._det_characteristic_points()
        self.acc = None
        if acc is not None:
            self.acc = acc

        self.data = {}

        self._create_stencil()

    def apply(self, u, idx0):
        """ Applies the stencil to a point in an equidistant grid.

        :param u: ndarray
            An array with the function to differentiate.

        :param idx0: int or tuple of ints
            The index of the grid point where to differentiate the function.

        :return:
            The derivative at the given point.
        """

        if not hasattr(idx0, '__len__'):
            idx0 = (idx0, )

        typ = []
        for axis in range(len(self.shape)):
            if idx0[axis] == 0:
                typ.append('L')
            elif idx0[axis] == self.shape[axis] - 1:
                typ.append('H')
            else:
                typ.append('C')
        typ = tuple(typ)

        stl = self.data[typ]

        idx0 = np.array(idx0)
        du = 0.
        for o, c in stl.items():
            idx = idx0 + o
            du += c * u[tuple(idx)]

        return du

    def apply_all(self, u):
        """ Applies the stencil to all grid points.

        :param u: ndarray
            An array with the function to differentiate.

        :return:
            An array with the derivative.
        """

        assert self.shape == u.shape

        ndims = len(u.shape)
        if ndims == 1:
            indices = list(range(len(u)))
        else:
            axes_indices = []
            for axis in range(ndims):
                axes_indices.append(list(range(u.shape[axis])))

            axes_indices = tuple(axes_indices)
            indices = list(product(*axes_indices))

        du = np.zeros_like(u)

        for idx in indices:
            du[idx] = self.apply(u, idx)

        return du

    def for_point(self, idx):
        """ The returns the stencil for a given grid point.

        Stencil forms are different depending on whether the grid point is in the interior
        or at some boundary. This function selects the appropriate one.

        :param idx: int or tuple of ints
            The index of the grid point.

        :return:
            The stencil data for that point.
        """
        typ = self.type_for_point(idx)
        return self.data[typ]

    def type_for_point(self, idx):
        typ = []
        for axis in range(len(idx)):
            if idx[axis] == 0:
                typ.append('L')
            elif idx[axis] == self.shape[axis] - 1:
                typ.append('H')
            else:
                typ.append('C')
        return tuple(typ)

    def _create_stencil(self):

        matrix = self.diff_op.matrix(self.shape, acc=self.acc)

        for pt in self.char_pts:

            char_point_stencil = {}
            self.data[pt] = char_point_stencil

            index_tuple_for_char_pt = self._typical_index_tuple_for_char_point(pt)
            long_index_for_char_pt = to_long_index(index_tuple_for_char_pt, self.shape)

            row = matrix[long_index_for_char_pt, :]
            long_row_inds, long_col_inds = row.nonzero()

            for long_offset_ind in long_col_inds:
                offset_ind_tuple = np.array(to_index_tuple(long_offset_ind, self.shape), dtype=np.int)
                offset_ind_tuple -= np.array(index_tuple_for_char_pt, dtype=np.int)
                char_point_stencil[tuple(offset_ind_tuple)] = row[0, long_offset_ind]

    def _typical_index_tuple_for_char_point(self, pt):
        index_tuple_for_char_pt = []
        for axis, key in enumerate(pt):
            if key == 'L':
                index_tuple_for_char_pt.append(0)
            elif key == 'C':
                index_tuple_for_char_pt.append(self.shape[axis] // 2)
            else:
                index_tuple_for_char_pt.append(self.shape[axis] - 1)
        return tuple(index_tuple_for_char_pt)

    def _det_characteristic_points(self):
        shape = self.shape
        ndim = len(shape)
        typ = [("L", "C", "H")]*ndim
        return product(*typ)

    def __str__(self):
        s = ""
        for typ, stl in self.data.items():
            s += str(typ) + ":\t" + str(stl) + "\n"
        return s




"""
This module determines finite difference coefficients for uniform and 
non-uniform grids for any desired accuracy order.
"""

import math
import numpy as np


def coefficients(deriv, acc):
    """
    Calculates the finite difference coefficients for given derivative order and accuracy order.
    Assumes that the underlying grid is uniform. This function is available at the `findiff`
    package level.
    
    :param deriv: int > 0: The derivative order.
          
    :param acc:  even int > 0: The accuracy order. 
          
    :return: dict with the finite difference coefficients and corresponding offsets 
    """

    if acc % 2 == 1:
        acc += 1

    ret = {}

    # Determine central coefficients

    num_central = 2 * math.floor((deriv + 1) / 2) - 1 + acc
    num_side = num_central // 2

    ret["center"] = _calc_coef(num_side, num_side, deriv)

    # Determine forward coefficients

    if deriv % 2 == 0:
        num_coef = num_central + 1
    else:
        num_coef = num_central

    ret["forward"] = _calc_coef(0, num_coef - 1, deriv)

    # Determine backward coefficients

    ret["backward"] = _calc_coef(num_coef - 1, 0, deriv)

    return ret


def _calc_coef(left, right, deriv):

    matrix = _build_matrix(left, right, deriv)

    rhs = _build_rhs(left, right, deriv)

    return {
        "coefficients": np.linalg.solve(matrix, rhs),
        "offsets": np.array([p for p in range(-left, right+1)])
    }


def coefficients_non_uni(deriv, acc, coords, idx):
    """
    Calculates the finite difference coefficients for given derivative order and accuracy order.
    Assumes that the underlying grid is non-uniform.

    :param deriv: int > 0: The derivative order.

    :param acc:  even int > 0: The accuracy order.
     
    :param coords:  1D numpy.ndarray: the coordinates of the axis for the partial derivative
    
    :param idx:  int: index of the grid position where to calculate the coefficients

    :return: dict with the finite difference coefficients and corresponding offsets 
    """

    if acc % 2 == 1:
        acc += 1

    num_central = 2 * math.floor((deriv + 1) / 2) - 1 + acc
    num_side = num_central // 2

    if deriv % 2 == 0:
        num_coef = num_central + 1
    else:
        num_coef = num_central

    if idx < num_side:
        matrix = _build_matrix_non_uniform(0, num_coef - 1, coords, idx)

        rhs = _build_rhs(0, num_coef - 1, deriv)

        ret = {
            "coefficients": np.linalg.solve(matrix, rhs),
            "offsets": np.array([p for p in range(num_coef)])
        }

    elif idx >= len(coords) - num_side:
        matrix = _build_matrix_non_uniform(num_coef - 1, 0, coords, idx)

        rhs = _build_rhs(num_coef - 1, 0, deriv)

        ret = {
            "coefficients": np.linalg.solve(matrix, rhs),
            "offsets": np.array([p for p in range(-num_coef + 1, 1)])
        }

    else:
        matrix = _build_matrix_non_uniform(num_side, num_side, coords, idx)

        rhs = _build_rhs(num_side, num_side, deriv)

        ret = {
            "coefficients": np.linalg.solve(matrix, rhs),
            "offsets": np.array([p for p in range(-num_side, num_side + 1)])
        }

    return ret


def _build_matrix(p, q, deriv):
    """Constructs the equation system matrix for the finite difference coefficients"""
    A = [([1 for _ in range(-p, q+1)])]
    for i in range(1, p + q + 1):
        A.append([j**i for j in range(-p, q+1)])
    return np.array(A,dtype='float')


def _build_rhs(p, q, deriv):
    """The right hand side of the equation system matrix"""

    b = [0 for _ in range(p+q+1)]
    b[deriv] = math.factorial(deriv)
    return np.array(b,dtype='float')


def _build_matrix_non_uniform(p, q, coords, k):
    """Constructs the equation matrix for the finite difference coefficients of non-uniform grids at location k"""
    A = [[1] * (p+q+1)]
    for i in range(1, p + q + 1):
        line = [(coords[k+j] - coords[k])**i for j in range(-p, q+1)]
        A.append(line)
    return np.array(A,dtype='float')



import sys
sys.path.insert(1, '..')

import unittest
from findiff import coefficients
from findiff.coefs import coefficients_non_uni

import numpy as np



class TestCoefs(unittest.TestCase):

    def test_order2_acc2(self):

        c = coefficients(deriv=2, acc=2)

        coefs = c["center"]["coefficients"]
        np.testing.assert_array_almost_equal(np.array([1., -2., 1.]), coefs)
        offs = c["center"]["offsets"]
        np.testing.assert_array_almost_equal(np.array([-1, 0, 1]), offs)

        coefs = c["forward"]["coefficients"]
        np.testing.assert_array_almost_equal(np.array([2, -5, 4, -1]), coefs)
        offs = c["forward"]["offsets"]
        np.testing.assert_array_almost_equal(np.array([0, 1, 2, 3]), offs)

        coefs = c["backward"]["coefficients"]
        np.testing.assert_array_almost_equal(np.array(([2, -5, 4, -1])[::-1]), coefs)
        offs = c["backward"]["offsets"]
        np.testing.assert_array_almost_equal(np.array([-3, -2, -1, 0]), offs)

    def test_order1_acc2(self):

        c = coefficients(deriv=1, acc=2)

        coefs = c["center"]["coefficients"]
        np.testing.assert_array_almost_equal(np.array([-0.5, 0, 0.5]), coefs)
        offs = c["center"]["offsets"]
        np.testing.assert_array_almost_equal(np.array([-1, 0, 1]), offs)

        coefs = c["forward"]["coefficients"]
        np.testing.assert_array_almost_equal(np.array([-1.5, 2, -0.5]), coefs)
        offs = c["forward"]["offsets"]
        np.testing.assert_array_almost_equal(np.array([0, 1, 2]), offs)

        coefs = c["backward"]["coefficients"]
        np.testing.assert_array_almost_equal(-np.array([-1.5, 2, -0.5])[::-1], coefs)
        offs = c["backward"]["offsets"]
        np.testing.assert_array_almost_equal(np.array([-2, -1, 0]), offs)

    def test_order1_acc4(self):

        c = coefficients(deriv=1, acc=4)
        coefs = c["center"]["coefficients"]
        np.testing.assert_array_almost_equal(np.array([1/12, -2/3, 0, 2/3, -1/12]), coefs)
        offs = c["center"]["offsets"]
        np.testing.assert_array_almost_equal(np.array([-2, -1, 0, 1, 2]), offs)

        coefs = c["forward"]["coefficients"]
        np.testing.assert_array_almost_equal(np.array([-25/12, 4, -3, 4/3, -1/4]), coefs)
        offs = c["forward"]["offsets"]
        np.testing.assert_array_almost_equal(np.array([0, 1, 2, 3, 4]), offs)

        coefs = c["backward"]["coefficients"]
        np.testing.assert_array_almost_equal(-np.array([-25/12, 4, -3, 4/3, -1/4])[::-1], coefs)
        offs = c["backward"]["offsets"]
        np.testing.assert_array_almost_equal(-np.array([0, 1, 2, 3, 4])[::-1], offs)

    def test_order2_acc4(self):

        c = coefficients(deriv=2, acc=4)
        coefs = c["center"]["coefficients"]
        np.testing.assert_array_almost_equal(np.array([-1/12, 4/3, -2.5, 4/3, -1/12]), coefs)
        offs = c["center"]["offsets"]
        np.testing.assert_array_almost_equal(np.array([-2, -1, 0, 1, 2]), offs)

        coefs = c["forward"]["coefficients"]
        np.testing.assert_array_almost_equal(np.array([15/4, -77/6, 107/6, -13, 61/12, -5/6]), coefs)
        offs = c["forward"]["offsets"]
        np.testing.assert_array_almost_equal(np.array([0, 1, 2, 3, 4, 5]), offs)

        coefs = c["backward"]["coefficients"]
        np.testing.assert_array_almost_equal(np.array([15/4, -77/6, 107/6, -13, 61/12, -5/6])[::-1], coefs)
        offs = c["backward"]["offsets"]
        np.testing.assert_array_almost_equal(-np.array([0, 1, 2, 3, 4, 5])[::-1], offs)

    def test_non_uniform(self):

        x = np.linspace(0, 10, 100)
        dx = x[1] - x[0]

        c_uni = coefficients(deriv=2, acc=2)
        coefs_uni = c_uni["center"]["coefficients"]/dx**2

        c_non_uni = coefficients_non_uni(deriv=2, acc=2, coords=x, idx=5)
        coefs_non_uni = c_non_uni["coefficients"]

        np.testing.assert_array_almost_equal(coefs_non_uni, coefs_uni)


if __name__ == '__main__':
    unittest.main()


import numpy as np
import scipy.sparse as sparse
from scipy.sparse.linalg import spsolve


class PDE(object):

    def __init__(self, lhs, rhs, bcs):
        self.lhs = lhs
        self.rhs = rhs
        self.bcs = bcs
        self._L = None

    def solve(self):

        shape = self.bcs.shape
        if self._L is None:
            self._L = self.lhs.matrix(shape) # expensive operation, so cache it

        L = sparse.lil_matrix(self._L)
        f = self.rhs.reshape(-1, 1)

        nz = list(self.bcs.row_inds())

        L[nz, :] = self.bcs.lhs[nz, :]
        f[nz] = np.array(self.bcs.rhs[nz].toarray()).reshape(-1, 1)

        L = sparse.csr_matrix(L)
        return spsolve(L, f).reshape(shape)


class BoundaryConditions(object):

    def __init__(self, shape):
        self.shape = shape
        siz = np.prod(shape)
        self.long_indices = np.array(list(range(siz))).reshape(shape)
        self.lhs = sparse.lil_matrix((siz, siz))
        self.rhs = sparse.lil_matrix((siz, 1))

    def __setitem__(self, key, value):

        lng_inds = self.long_indices[key]

        if isinstance(value, tuple): # Neumann BC
            op, value = value
            # Avoid calling matrix for the whole grid! Optimize later!
            mat = sparse.lil_matrix(op.matrix(self.shape))
            self.lhs[lng_inds, :] = mat[lng_inds, :]
        else: # Dirichlet BC
            self.lhs[lng_inds, lng_inds] = 1

        if isinstance(value, np.ndarray):
            value = value.reshape(-1)[lng_inds]
            for i, v in zip(lng_inds, value):
                self.rhs[i] = v
        else:
            self.rhs[lng_inds] = value

    def row_inds(self):
        nz_rows, nz_cols = self.lhs.nonzero()
        return nz_rows
